/* vi: set ts=2 shiftwidth=2 expandtab:
 *
 * Copyright (C) 2006-2007  Simon Baldwin (simon_baldwin@yahoo.com)
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307
 * USA
 */

#include <assert.h>

#include <glk.h>

#include "protos.h"


/*
 * hash()
 *
 * Hash a string, hashpjw algorithm, from 'Compilers, principles, techniques,
 * and tools', page 436, unmodulo'ed.  Used to create a strings digest to help
 * with comparisons.
 */
glui32
hash (const char *string)
{
  const char *sp;
  glui32 h = 0, g;

  for (sp = string; *sp != '\0'; sp++)
      {
        h = (h << 4) + (*sp);
        if ((g = (h & 0xF0000000)) != 0)
          {
            h = h ^ (g >> 24);
            h = h ^ g;
          }
      }
  return h;
}
